/*
 * Copyright (C) 2004 Lee Willis <lee@leewillis.co.uk>
 *           (C) 2005 Jorn Baayen <jbaayen@gnome.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using System;
using System.IO;
using System.Collections;
using System.Net.Sockets;
using System.Xml;
using System.Text;
using System.Runtime.InteropServices;

using Gtk;

using Mono.Posix;

using Muine.PluginLib;

namespace Muine.Plugins.Burn {
    
    public class BurnPlugin : Plugin
    {
        [DllImport ("libmuineburn")]
        private static extern ulong get_dir_free_space (string path);

        private IPlayer player;
        private Recorder recorder;

        public override void Initialize (IPlayer player)
        {
            this.player = player;
            ActionGroup action_group;
        
            ActionEntry [] action_entries = new ActionEntry [] {
                new ActionEntry ("Burn", Gtk.Stock.Execute, 
                                 "_Burn", "<control><shift>B", null, new EventHandler (OnBurn))
            };

            action_group = new ActionGroup ("BurnPluginActions");
            action_group.Add (action_entries);

            player.UIManager.InsertActionGroup (action_group, -1);

            player.UIManager.AddUi (player.UIManager.NewMergeId (),
                                    "/MenuBar/FileMenu/ExtraFileActions", "BurnMenuItem",
                                    "Burn", UIManagerItemType.Menuitem, false);

        }
    
        private void OnBurn (object o, System.EventArgs events)
        {
            string temp_dir;
            string drive_path;

            try {
                recorder = new Recorder ();
            } catch (Exception e) {
                new ErrorDialog (string.Format (Catalog.GetString ("Failed to initialize the recorder backend:\n{0}\n\nExiting..."), e.Message));
                return;
            }

            ISong [] songs = player.Playlist;
            foreach (ISong song in songs) {
                recorder.AddSong (song);
                Console.WriteLine (song.Title);
            }

            temp_dir = GetBurnTempDir ();
            if (temp_dir == null)
                return;

            drive_path = GetBurnDrivePath ();
            if (drive_path == null)
                return;

            recorder.InsertCdEvent += new Recorder.InsertCdHandler (HandleInsertCdEvent);

            recorder.WriteFiles (temp_dir);

        }


        private bool HandleInsertCdEvent (bool is_reload, bool can_rewrite, bool busy_cd)
        {
            string msg;
            string title;
            if (busy_cd) {
                msg = Catalog.GetString ("Please make sure another application is not using the disc.");
                title = Catalog.GetString ("Disc is busy");
            } else if (is_reload && can_rewrite) {
                msg = Catalog.GetString ("Please insert a rewritable or blank media in the drive tray.");
                title = Catalog.GetString ("Insert rewritable or blank media");
            } else if (is_reload && !can_rewrite) {
                msg = Catalog.GetString ("Please insert a blank media in the drive tray.");
                title = Catalog.GetString ("Insert blank media");
            } else if (can_rewrite) {
                msg = Catalog.GetString ("Please replace the in-drive media by a rewritable or blank media.");
                title = Catalog.GetString ("Reload rewritable or blank media");
            } else {
                msg = Catalog.GetString ("Please replace the in-drive media by a blank media.");
                title = Catalog.GetString ("Reload blank media");
            }

            Dialog dialog = new Dialog (title, player.Window, DialogFlags.Modal);
            dialog.AddButton (Stock.Ok, ResponseType.Ok);
            dialog.AddButton (Stock.Cancel, ResponseType.Cancel);
            Label label = new Label (msg);
            label.Visible = true;
            dialog.VBox.PackStart (label);
            if (dialog.Run () == (int) ResponseType.Cancel) {
                dialog.Destroy ();
                return false;
            }

            dialog.Destroy ();
            return true;
        }



        private string GetBurnTempDir ()
        {
            string dir = Path.GetTempFileName ();

            if (File.Exists (dir))
                File.Delete (dir);

            Directory.CreateDirectory (dir);
            return dir;
        }

        private string GetBurnDrivePath ()
        {
            string drive_path;

            try {
                drive_path = (string) Config.Get ("/apps/muine/burn/drive");
            } catch {
                drive_path = null;
            }
            if (drive_path != null) {
                if (recorder.SetDrive (drive_path) == false)
                    drive_path = null;
            } 

            ArrayList devices = recorder.GetValidDevices ();
            if (devices.Count == 0) {
                new ErrorDialog (string.Format (Catalog.GetString ("You don't have any recognized CD Recorder")));
                return null;
            }
                
            if (drive_path == null && devices.Count == 1) {
                CdUnit unit = (CdUnit) devices[0];
                recorder.SetDrive (unit.Path);
            }

            if (drive_path == null || devices.Count > 1) {
                BurnPrefs bp = new BurnPrefs (devices);
                if (bp.GetDrivePath (out drive_path) == false)
                    return null;
                if (recorder.SetDrive (drive_path) == false)
                    return null;
            } 
            return drive_path;
        }

    }

}
