/*
 * Copyright (C) 2004 Jorn Baayen <jorn@nl.linux.org>
 * Copyright (C) 2004 Fernando Herrera <fherrera@onirica.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


using Gtk;
using System;
using System.Runtime.InteropServices;
using System.Collections;
using System.IO;

using Mono.Posix;

using Muine.PluginLib;

namespace Muine.Plugins.Burn {

    public class Recorder : GLib.Object
    {
        [DllImport ("libmuineburn")]
        private static extern bool recorder_set_orig (IntPtr recorder,
                                                      string filename,
                                                      out IntPtr error_ptr);

        [DllImport ("libmuineburn")]
        private static extern bool recorder_set_dest (IntPtr recorder,
                                                      string filename,
                                                      out IntPtr error_ptr);

        [DllImport ("libmuineburn")]
        private static extern bool recorder_add_track (IntPtr recorder,
                                                       string filename,
                                                       out IntPtr error_ptr);
        [DllImport ("libmuineburn")]
        private static extern bool recorder_has_many_devices ();

        [DllImport ("libmuineburn")]
        private static extern string recorder_get_valid_devices ();

        [DllImport ("libmuineburn")]
        private static extern bool recorder_select_drive (IntPtr recorder,
                                                          string device,
                                                          out IntPtr error_ptr);
        [DllImport ("libmuineburn")]
        private static extern long recorder_get_media_size (IntPtr recorder);

        [DllImport ("libmuineburn")]
        private static extern int recorder_burn_cd (IntPtr recorder);

        private bool writing = false;
        private int num_songs = 0;
        private ulong duration = 0;
        private string temp_dir = null;
        ProgressWindow pw;

        private Queue songs;
        private Queue files;

        public void AddSong (ISong song) {
            songs.Enqueue (song);
            duration += (ulong) song.Duration;
        }

        public ulong GetDuration () {
            return duration;
        }

        [DllImport ("libmuineburn")]
        private static extern void recorder_write_file (IntPtr recorder);


        [DllImport ("libmuineburn")]
        private static extern void recorder_write_file_cancel (IntPtr recorder);

        private void WriteFileCancel ()
        {
            recorder_write_file_cancel (Raw);
        }

        private void WriteFile ()
        {    
            IntPtr error_ptr;
            ISong song;

            if (songs.Count == 0)
                return;

            song = (ISong) songs.Dequeue ();

            if (pw.ReportFile (song.Title) == false) {
                WriteFileCancel ();
                return;
            }
        
            recorder_set_orig (Raw, song.Filename, out error_ptr);
            if (error_ptr != IntPtr.Zero) {
                string error = GLib.Marshaller.PtrToStringGFree (error_ptr);

                new ErrorDialog (string.Format (Catalog.GetString ("Error opening {0}:\n{1}"),
                                                song, error));
            }
            
            recorder_set_dest (Raw, temp_dir + System.IO.Path.GetFileNameWithoutExtension (song.Filename) + ".wav",
                               out error_ptr);
            if (error_ptr != IntPtr.Zero) {
                string error = GLib.Marshaller.PtrToStringGFree (error_ptr);

                new ErrorDialog (string.Format (Catalog.GetString ("Error opening {0}{1}.wav:\n{2}"),
                                                temp_dir,
                                                System.IO.Path.GetFileNameWithoutExtension (song.Filename),
                                                error));
            }

            recorder_write_file (Raw);
            recorder_add_track (Raw, temp_dir + System.IO.Path.GetFileNameWithoutExtension (song.Filename) + ".wav",
                                out error_ptr);

            files.Enqueue (temp_dir + System.IO.Path.GetFileNameWithoutExtension (song.Filename) + ".wav");

            if (error_ptr != IntPtr.Zero) {
                string error = GLib.Marshaller.PtrToStringGFree (error_ptr);

                new ErrorDialog (string.Format (Catalog.GetString ("Error opening {0}:\n{1}"),
                                                System.IO.Path.GetFileNameWithoutExtension (song.Filename),
                                                error));
            }
        }

        public void WriteFiles (string path)
        {
            temp_dir = path;
            num_songs = songs.Count;
            if (num_songs >= 1) {
                pw = new ProgressWindow (null, Catalog.GetString ("Preparing Files:"), 
                                         Catalog.GetString ("Burning CD"));
                pw.TotalSteps = num_songs;
                WriteFile ();
            }
        }

        public ArrayList GetValidDevices ()
        {
            ArrayList res = new ArrayList();
            string str = recorder_get_valid_devices ();

            int pos = 0;
            int end = str.IndexOf (';', pos);
            Console.WriteLine (str);
            while (end > 0) {
                int p = str.IndexOf (':', pos);
                CdUnit unit = new CdUnit (str.Substring (pos, p - pos),
                                          str.Substring (p + 1, end - p - 1 ));
                pos = end + 1;
                res.Add (unit);
                if (pos >= str.Length)
                    break;
                end = str.IndexOf (';', pos);
            }
            return res;
            
        }


        public bool HasManyDevices ()
        {
            return recorder_has_many_devices ();
        }

        public bool SetDrive (string path)
        {
            IntPtr error_ptr;
            recorder_select_drive (Raw, path, out error_ptr);
            if (error_ptr != IntPtr.Zero) {
                string error = GLib.Marshaller.PtrToStringGFree (error_ptr);

                new ErrorDialog (string.Format (Catalog.GetString ("Error selection drive:\n{0}"),
                                                error));
                return false;
            }

            return true;
        }

        public void BurnCd ()
        {
            recorder_burn_cd (Raw);
            /* FIXME Show error or ok */

            string file;
            do {
                file = (string) files.Dequeue ();
                File.Delete (file);
            } while (files.Count >= 1);
        
            pw.Done ();
        }



        public delegate bool InsertCdHandler (bool is_reload, bool can_rewrite, bool busy_cd);
        public event InsertCdHandler InsertCdEvent;


        [DllImport ("libmuineburn")]
        private static extern IntPtr recorder_new (out IntPtr error_ptr);

        private class Connect
        {
            [DllImport ("libgobject-2.0-0.dll")]
            public static extern uint g_signal_connect_data (IntPtr obj, string name,
                                                             SignalDelegate cb, IntPtr data,
                                                             IntPtr p, int flags);
        }

        private class ConnectInt
        {
            [DllImport ("libgobject-2.0-0.dll")]
            public static extern uint g_signal_connect_data (IntPtr obj, string name,
                                                             IntSignalDelegate cb, IntPtr data,
                                                             IntPtr p, int flags);
        }

        private class Connectstring
        {
            [DllImport ("libgobject-2.0-0.dll")]
            public static extern uint g_signal_connect_data (IntPtr obj, string name,
                                                             stringSignalDelegate cb, IntPtr data,
                                                             IntPtr p, int flags);
        }

        private class ConnectDouble
        {
            [DllImport ("libgobject-2.0-0.dll")]
            public static extern uint g_signal_connect_data (IntPtr obj, string name,
                                                             DoubleSignalDelegate cb, IntPtr data,
                                                             IntPtr p, int flags);
        }

        private class ConnectBoolBoolBool
        {
            [DllImport ("libgobject-2.0-0.dll")]
            public static extern uint g_signal_connect_data (IntPtr obj, string name,
                                                             BoolBoolBoolSignalDelegate cb, IntPtr data,
                                                             IntPtr p, int flags);
        }

        public Recorder () : base (IntPtr.Zero)
        {
            IntPtr error_ptr;
        
            songs = new Queue ();
            files = new Queue ();
            Raw = recorder_new (out error_ptr);
            if (error_ptr != IntPtr.Zero) {
                string error = GLib.Marshaller.PtrToStringGFree (error_ptr);
                throw new Exception (error);
            }
        
            Connect.g_signal_connect_data (Raw, "end_of_stream", new SignalDelegate (EosCallback),
                                           IntPtr.Zero, IntPtr.Zero, 0);
            Connectstring.g_signal_connect_data (Raw, "error", new stringSignalDelegate (ErrorCallback),
                                                 IntPtr.Zero, IntPtr.Zero, 0);
            ConnectDouble.g_signal_connect_data (Raw, "cd_burn_progress", new DoubleSignalDelegate (CdProgressCallback),
                                                 IntPtr.Zero, IntPtr.Zero, 0);

            ConnectBoolBoolBool.g_signal_connect_data (Raw, "cd_burn_insert_cd", new BoolBoolBoolSignalDelegate (CdInsertionCallback),
                                                       IntPtr.Zero, IntPtr.Zero, 0);



        }

        ~Recorder ()
        {
            Dispose ();
        }

        private delegate void SignalDelegate (IntPtr obj);
        private delegate void IntSignalDelegate (IntPtr obj, int i);
        private delegate void stringSignalDelegate (IntPtr obj, string s);
        private delegate void DoubleSignalDelegate (IntPtr obj, double d);
        private delegate bool BoolBoolBoolSignalDelegate (IntPtr obj, bool b1, bool b2, bool b3);

        public delegate void EndOfStreamEventHandler ();
        public event EndOfStreamEventHandler EndOfStreamEvent;

        private void EosCallback (IntPtr obj)
        {
            long media_size = 0;
            if (EndOfStreamEvent != null)
                EndOfStreamEvent ();
            if (songs.Count > 0) {
                pw.Step ();
                WriteFile ();
            } else {
                do {
                    string title, msg;
                    Dialog dialog;
                    int res;

                    media_size =  recorder_get_media_size (Raw);
                    if (media_size >= (long) duration)
                        break;
                    if (media_size == -1 /* CD_MEDIA_SIZE_UNKNOWN */)
                        break;
                    if (media_size == -3 /* CD_MEDIA_SIZE_BUSY */)
                        continue;
                    if (media_size == -2 /* CD_MEDIA_SIZE_NA */) {
                        msg = string.Format (Catalog.GetString ("Please insert a blank disc with at least {0} minutes free"), (duration / 60));
                        title = Catalog.GetString ("Insert Disc");
                        dialog = new Dialog (title, pw.window, DialogFlags.Modal);
                        dialog.AddButton (Stock.Ok, ResponseType.Ok);
                        dialog.AddButton (Stock.Cancel, ResponseType.Cancel);
                    } else {
                    
                        dialog = new Dialog ();
                        msg = string.Format (Catalog.GetString ("Please replace the in-drive media with a blank disc with at least {0} minutes free"), (duration / 60));
                        title = Catalog.GetString ("Replace Disc");
                        dialog = new Dialog (title, pw.window, DialogFlags.Modal);
                        dialog.AddButton (Catalog.GetString ("_Continue"), ResponseType.Ok);
                        dialog.AddButton (Stock.Cancel, ResponseType.Cancel);
                    }

                    Label label = new Label (msg);
                    label.Visible = true;
                    dialog.VBox.PackStart (label);
                    
                    res = dialog.Run ();
                    dialog.Destroy ();
                    if (res == (int) ResponseType.Cancel) {
                        pw.Done ();
                        return ;
                    }
                    if (res == (int) ResponseType.Yes) {
                        //overburn = true;
                        break;
                    }
                } while (media_size < 0 || (long) duration > ((media_size / 1024 / 1024) - 1) * 48 / 7 / 60);


                pw.Label = Catalog.GetString ("Burning CD:");
                pw.ReportFile (Catalog.GetString ("Initializing Drive"));
                BurnCd ();
            }
        }

        private void ErrorCallback (IntPtr obj, string error)
        {
            pw.Done ();
            new ErrorDialog (string.Format (Catalog.GetString ("Recorder backend error:\n{0}"), error));
            Dispose ();
        }
        private void CdProgressCallback (IntPtr obj, double fraction)
        {
            pw.Fraction = fraction;
            pw.ReportFile (Catalog.GetString ("Progress: " + (int) (fraction * 100) + "%"));
        }
        private bool CdInsertionCallback (IntPtr obj, bool is_reload, bool can_rewrite, bool busy_cd)
        {
            return InsertCdEvent (is_reload, can_rewrite, busy_cd);
        }
    
    
    }
}
