/* -*- mode: C; c-file-style: "gnu" -*- */
/*
 * Copyright (C) 2004 Fernando Herrera <fherrera@onirica.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <string.h>
#include <math.h>
#include <gst/gst.h>
#include <gst/gsterror.h>
#include <gst/gconf/gconf.h>
#include <gst/play/play.h>
#include <glib/gi18n.h>
#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-file-size.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <nautilus-burn-drive.h>
#include <nautilus-burn-recorder.h>
#include <sys/statvfs.h>

#include "recorder.h"
#include "recorder-marshal.h"

static void recorder_class_init (RecorderClass     *klass);
static void recorder_init       (Recorder          *recorder);
static void recorder_finalize   (GObject         *object);
static void eos_cb            (GstElement      *sink,
		               Recorder          *recorder);
static void error_cb (GstElement *gstelement, GstElement *arg1, GError *arg2, gchar *arg3, Recorder *recorder);

static void recorder_cd_progress_cb (NautilusBurnRecorder *cdrecorder,
				     gdouble fraction,
				     gpointer data);

static gboolean recorder_insert_cd_request_cb (NautilusBurnRecorder *cdrecorder,
					       gboolean is_reload,
					       gboolean can_rewrite,
					       gboolean busy_cd,
					       gpointer data);



enum {
	END_OF_STREAM,
	CD_BURN_PROGRESS,
	CD_BURN_INSERT_CD,
	ERROR,
	LAST_SIGNAL
};

struct _RecorderPriv {
	GstElement *pipeline;
	GstElement *source;
	GstElement *decoder;
	GstElement *writer;
	GstElement *converter;
	GstElement *encoder;
	GstElement *sink;

	guint	    eos_idle_id;
	gulong      eos_idle_handler;

	gint64      pos;

	GList	   *tracks;
	NautilusBurnDrive	   *drive;
	NautilusBurnRecorder *cdrecorder;
};

static GObjectClass *parent_class;
static guint signals[LAST_SIGNAL];

GType
recorder_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (RecorderClass),
				NULL,           /* base_init */
				NULL,           /* base_finalize */
				(GClassInitFunc) recorder_class_init,
				NULL,           /* class_finalize */
				NULL,           /* class_data */
				sizeof (Recorder),
				0,
				(GInstanceInitFunc) recorder_init,
			};

			type = g_type_register_static (G_TYPE_OBJECT,
						       "Recorder",
				                       &info, 0);
	}

	return type;
}

static void
recorder_class_init (RecorderClass *klass)
{
	GObjectClass *object_class;

	parent_class = g_type_class_peek_parent (klass);
	object_class = (GObjectClass *) klass;

	object_class->finalize = recorder_finalize;

	signals[END_OF_STREAM] =
		g_signal_new ("end_of_stream",
		              G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	signals[CD_BURN_PROGRESS] =
		g_signal_new ("cd_burn_progress",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      g_cclosure_marshal_VOID__DOUBLE,
			      G_TYPE_NONE, 1, G_TYPE_DOUBLE);
	signals[CD_BURN_INSERT_CD] =
		g_signal_new ("cd_burn_insert_cd",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      recorder_marshal_BOOLEAN__BOOLEAN_BOOLEAN_BOOLEAN,
			      G_TYPE_BOOLEAN, 3, G_TYPE_BOOLEAN,
			      G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
	signals[ERROR] =
		g_signal_new ("error",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      0,
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_STRING);
}

static void
recorder_init (Recorder *recorder)
{
}

static void
recorder_construct (Recorder *recorder, char **error)
{
	RecorderPriv *priv;

	gst_init (NULL, NULL);

	priv = g_new0 (RecorderPriv, 1);
	recorder->priv = priv;

	priv->eos_idle_id = 0;

	priv->pipeline = gst_pipeline_new ("pipeline");

	priv->source = gst_element_factory_make ("gnomevfssrc", "source");
	if (!priv->source) {
		*error = g_strdup (_("Failed to create the required gnomevfssrc GStreamer element"));

		return;
	}

	priv->decoder = gst_element_factory_make ("mad", "decoder");
	if (!priv->decoder) {
		*error = g_strdup (_("Could not render GStreamer mad decoder"));

		return;
	}

	priv->converter = gst_element_factory_make ("audioconvert", "converter");

	priv->encoder = gst_element_factory_make ("wavenc", "encoder");

	priv->sink = gst_element_factory_make ("filesink", "sink");

	gst_bin_add_many (GST_BIN (priv->pipeline), priv->source, priv->decoder,  priv->converter, priv->encoder, priv->sink, NULL);
	gst_element_link_many (priv->source, priv->decoder, priv->converter, priv->encoder, priv->sink, NULL);

	g_signal_connect (priv->pipeline, "error", G_CALLBACK (error_cb), recorder);
	priv->eos_idle_handler = g_signal_connect (priv->pipeline, "eos", G_CALLBACK (eos_cb), recorder);

	priv->tracks = NULL;
	priv->drive = NULL;

	priv->cdrecorder = nautilus_burn_recorder_new ();

	g_signal_connect (G_OBJECT (priv->cdrecorder), "progress-changed", G_CALLBACK (recorder_cd_progress_cb), recorder);
	g_signal_connect (G_OBJECT (priv->cdrecorder), "insert-media-request", G_CALLBACK (recorder_insert_cd_request_cb), recorder);
}

static void
recorder_finalize (GObject *object)
{
	Recorder *recorder = RECORDER (object);

	//recorder_write_file_cancel (recorder);

	g_free (recorder->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

Recorder *
recorder_new (char **error)
{
	Recorder *recorder;

	recorder = g_object_new (TYPE_RECORDER, NULL);

	*error = NULL;

	recorder_construct (recorder, error);

	return recorder;
}

static void
eos_cb (GstElement *sink, Recorder *recorder)
{
	recorder->priv->eos_idle_id = 0;
	gst_element_set_state (GST_ELEMENT (recorder->priv->pipeline), GST_STATE_NULL);
	g_signal_emit (recorder, signals[END_OF_STREAM], 0);
}

typedef struct {
	Recorder *recorder;
	char *error;
} RecorderError;

static gboolean
error_idle_cb (RecorderError *data)
{
	g_signal_emit (data->recorder, signals[ERROR], 0, data->error);

	g_free (data->error);
	g_free (data);

	return FALSE;
}

static void
error_cb (GstElement *gstelement,
          GstElement *arg1,
          GError *error,
          gchar *str,
	  Recorder    *recorder)
{
	RecorderError *data = g_new0 (RecorderError, 1);

	data->recorder = recorder;
	data->error = g_strconcat (error->message, ":\n", str, NULL);

	g_signal_handler_disconnect (recorder->priv->pipeline, recorder->priv->eos_idle_handler);

	g_idle_add ((GSourceFunc) error_idle_cb, data);
}

gboolean
recorder_set_orig (Recorder     *recorder,
		   const char *file,
		   char      **error)
{
	g_return_val_if_fail (IS_RECORDER (recorder), FALSE);

	*error = NULL;

	if (recorder->priv->eos_idle_id > 0) {
		g_source_remove (recorder->priv->eos_idle_id);
		recorder->priv->eos_idle_id = 0;
	}

	g_object_set (G_OBJECT (recorder->priv->source), "location", file, NULL);

	recorder->priv->pos = 0;

	return TRUE;
}

gboolean
recorder_set_dest (Recorder     *recorder,
		   const char *file,
		   char      **error)
{
	g_return_val_if_fail (IS_RECORDER (recorder), FALSE);

	*error = NULL;

	if (recorder->priv->eos_idle_id > 0) {
		g_source_remove (recorder->priv->eos_idle_id);
		recorder->priv->eos_idle_id = 0;
	}

	g_object_set (G_OBJECT (recorder->priv->sink), "location", file, NULL);

	recorder->priv->pos = 0;

	return TRUE;
}


void
recorder_write_file (Recorder *recorder)
{
	g_return_if_fail (IS_RECORDER (recorder));

	gst_element_set_state (GST_ELEMENT (recorder->priv->pipeline), GST_STATE_PLAYING);

	recorder->priv->eos_idle_id = g_idle_add ((GSourceFunc) gst_bin_iterate, GST_BIN (recorder->priv->pipeline));

}

void
recorder_write_file_cancel (Recorder *recorder)
{
	g_return_if_fail (IS_RECORDER (recorder));

	recorder->priv->pos = 0;

	gst_element_set_state (GST_ELEMENT (recorder->priv->pipeline), GST_STATE_READY);
}

void
recorder_pause (Recorder *recorder)
{
	g_return_if_fail (IS_RECORDER (recorder));

	gst_element_set_state (GST_ELEMENT (recorder->priv->pipeline), GST_STATE_PAUSED);
}

int
recorder_tell (Recorder *recorder)
{
	g_return_val_if_fail (IS_RECORDER (recorder), -1);

	return recorder->priv->pos / GST_SECOND;
}

gboolean
recorder_add_track (Recorder     *recorder,
		   const char *file,
		   char      **error)
{
	NautilusBurnRecorderTrack *track;
	g_return_val_if_fail (IS_RECORDER (recorder), FALSE);

	*error = NULL;

	track = g_new0 (NautilusBurnRecorderTrack, 1);

	track->type = NAUTILUS_BURN_RECORDER_TRACK_TYPE_AUDIO;
	track->contents.audio.filename = g_strdup (file);
	track->contents.audio.cdtext = g_path_get_basename (file);


	/* FIXME: check for existence */

	recorder->priv->tracks = g_list_append (recorder->priv->tracks, track);

	return TRUE;
}

gboolean
recorder_has_many_devices ()
{
	GList *drives, *tmp;
	int num_drives = 0;

	drives = nautilus_burn_drive_get_list (FALSE, FALSE);

	for (tmp = drives; tmp != NULL; tmp = tmp->next) {
		NautilusBurnDrive *drive = (NautilusBurnDrive*) tmp->data;

		if (drive->type & (NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER |
				   NAUTILUS_BURN_DRIVE_TYPE_DVD_RAM_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER |
				   NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER))
			num_drives++;

		nautilus_burn_drive_free (drive);
	}
	g_list_free (drives);

	return num_drives > 1;
}

gchar*
recorder_get_valid_devices (void)
{
	GList *drives, *tmp;
	GString *buf;
	gchar *str;

	buf = g_string_new (NULL);

	drives = nautilus_burn_drive_get_list (FALSE, FALSE);

	for (tmp = drives; tmp != NULL; tmp = tmp->next) {
		NautilusBurnDrive *drive = (NautilusBurnDrive*) tmp->data;

		if (drive->type & (NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER |
				   NAUTILUS_BURN_DRIVE_TYPE_DVD_RAM_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER |
				   NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER)) {
			g_string_append (buf, drive->device);
			g_string_append (buf, ":");
			g_string_append (buf, drive->display_name);
			g_string_append (buf, ";");
		} 
		nautilus_burn_drive_free (drive);
	}
	g_list_free (drives);

	str = g_string_free (buf, FALSE);

	return str;
}

gboolean
recorder_select_drive  (Recorder     *recorder,
		        const char *device,
		        char      **error)
{
	GList *drives, *tmp;
	g_return_val_if_fail (IS_RECORDER (recorder), FALSE);

	*error = NULL;

	drives = nautilus_burn_drive_get_list (FALSE, FALSE);
	
	for (tmp = drives; tmp != NULL; tmp = tmp->next) {
		NautilusBurnDrive *drive = (NautilusBurnDrive*) tmp->data;
		if (strcmp (drive->device, device) == 0) {
			recorder->priv->drive = drive;
			break;
		}
		nautilus_burn_drive_free (drive);
	}
	g_list_free (drives);

	if (recorder->priv->drive == NULL) {
		*error = g_strdup_printf (_("Cannot found drive %s"), device);
		return FALSE;
	}

	if (!(recorder->priv->drive->type & (NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER |
					     NAUTILUS_BURN_DRIVE_TYPE_DVD_RAM_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER |
					     NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER | NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER))) {
		*error = g_strdup_printf (_("Drive %s cannot burn CDs"), device);
		return FALSE;
	}

	return TRUE;
}

gint64
recorder_get_media_size (Recorder *recorder)
{
	g_return_val_if_fail (IS_RECORDER (recorder), NAUTILUS_BURN_MEDIA_SIZE_NA);
	return nautilus_burn_drive_get_media_size (recorder->priv->drive);
}


/* FIXME: Exportar una funcion nueva para hacer get_size del CD */

int
recorder_burn_cd  (Recorder *recorder)
{
	g_return_val_if_fail (IS_RECORDER (recorder), FALSE);
	return nautilus_burn_recorder_write_tracks (recorder->priv->cdrecorder, recorder->priv->drive, recorder->priv->tracks, 0, 
                                                    NAUTILUS_BURN_RECORDER_WRITE_DISC_AT_ONCE |
                                                    NAUTILUS_BURN_RECORDER_WRITE_DEBUG);
}

static void
recorder_cd_progress_cb (NautilusBurnRecorder *cdrecorder, gdouble fraction, gpointer data)
{
	Recorder *recorder = (Recorder*) data;

	g_signal_emit (recorder, signals[CD_BURN_PROGRESS], 0, fraction);
}

static gboolean
recorder_insert_cd_request_cb (NautilusBurnRecorder *cdrecorder, gboolean is_reload, gboolean can_rewrite, gboolean busy_cd, gpointer data)
{

	Recorder *recorder = (Recorder*) data;
	gboolean res;

	g_signal_emit (recorder, signals[CD_BURN_INSERT_CD], 0, is_reload, can_rewrite, busy_cd, &res);
	return res;
}




unsigned long long
get_dir_free_space (char *path)
{
	GnomeVFSFileSize size;
	GnomeVFSURI *uri;

	uri = gnome_vfs_uri_new (path);
	if (gnome_vfs_get_volume_free_space (uri, &size) == GNOME_VFS_OK) {
		gnome_vfs_uri_unref (uri);
		return (unsigned long) size;
	} else {
		gnome_vfs_uri_unref (uri);
		return 0;
	}
}

